"""
Module for specification of the BRAIN file format API.

"""
import warnings

import numpy as np
import json
import os
import h5py

from brain.dataformat.base import ManagedFile, ManagedGroup, ManagedObject, ManagedObjectFile
from brain.dataformat.annotation import AnnotationDataGroup, Annotation, DataSelection


################################################
# ManagedFile                         #
################################################
class BrainDataFile(ManagedFile):
    """
    Class for management of HDF5 brain files.

    :ivar hdf_object: See ManagedFile

    Implicit instance variables (i.e., these are mapped names but not stored explicitly)

    :ivar data_#: Same as self.data(#). Usually use data_0
    :ivar descriptors_#: Same as self.descriptors(#). Usually use descriptors_0

    """
    def __init__(self, hdf_object, mode='r'):
        """
        :param hdf_object: This can be either the h5py.File object, a h5py.Group or h5py.Dataset
                           contained in the file of interest, or a string indicating the name
                           the file to be opened.
        :param mode: Used only if hdf_object is a string and a file is opened anew. Indicate
                     the mode in which a file should be opened, e.g., 'r', 'w', 'a'. See
                     the h5py.File documentation for details.
        """
        super(BrainDataFile, self).__init__(hdf_object=hdf_object, mode=mode)

    @classmethod
    def get_format_specification(cls):
        """
        Get dictionary describing the format.
        """
        return {'datasets': {},
                'groups': {},
                'managed_objects': [{'format_type': 'BrainDataData', 'optional': False},
                                    {'format_type': 'BrainDataDescriptors', 'optional': False}],
                'attributes': [],
                'group': None,
                'prefix': "entry_",
                'file_prefix': None,
                'file_extension': '.h5',
                'optional': False,
                'description': 'Managed BRAIN file.'}

    def populate(self, **kwargs):
        """
        Populate the Brain file with the Data and Descriptors group.
        """
        parent_object = self.hdf_object['/']
        BrainDataData.create(parent_object=parent_object)
        BrainDataDescriptors.create(parent_object=parent_object)
        self.flush()

    def __getitem__(self, item):
        """
        Enable slicing into the file
        """
        hdfobj = self.hdf_object[item]
        manobj = self.get_managed_object(hdfobj)
        if manobj is not None:
            return manobj
        else:
            return hdfobj

    def data(self, index=0):
        """
        Get the data managed object with the data. In
        principal the current format defines that there should be always
        exactly one descriptors group. However, multiple could be supported
        so this function is prepared to handle this in case the format should
        change in the future.

        :param index: Optional input parameter (should be always 0 for now)
                      to define which data group should be retrieved.
                      Currently their should be always exactly one.

        :returns: BrainDataData object or None
        """
        dataobjs = BrainDataData.get_all(self.hdf_object['/'])
        if len(dataobjs) == 0:
            return None
        else:
            return BrainDataData(dataobjs[index])

    def descriptors(self, index=0):
        """
        Get the descriptors managed object with the descriptors. In
        principal the current format defines that there should be always
        exactly one descriptors group. However, multiple could be supported
        so this function is prepared to handle this in case the format should
        change in the future.

        :param index: Optional input parameter (should be always 0 for now)
                      to define which descriptors group should be retrieved.
                      Currently their should be always exactly one.

        :returns: BrainDataDescriptors object or None
        """
        descriptorobjs = BrainDataDescriptors.get_all(self.hdf_object['/'])
        if len(descriptorobjs) == 0:
            return None
        else:
            return BrainDataDescriptors(descriptorobjs[index])

    def __getattr__(self, item):
        if item.startswith('data_'):
            indexstr = item.lstrip('data_')
            if indexstr.isdigit():
                return self.data(int(indexstr))
            else:
                raise AttributeError("'BrainDataFile' has not attribute" + item)
        elif item.startswith('descriptors_'):
            indexstr = item.lstrip('descriptors_')
            if indexstr.isdigit():
                return self.descriptors(int(indexstr))
            else:
                raise AttributeError("'BrainDataFile' has not attribute" + item)
        else:
            try:
                return super(BrainDataFile, self).__getattribute__(item)
            except AttributeError:
                return super(BrainDataFile, self).__getattr__(item)


################################################
# BrainDataMultiFile                           #
################################################
class BrainDataMultiFile(ManagedObjectFile):
    """
    Container file used to store a collection of BrainDataFile collections of neural data. This container is
    used to organize multiple  experiments, sessions, etc. stored in separate files into a single collection,
    allow a user to open a single file and interact with the collection of files as if they were all stored
    in the same file.

    NOTE: This container typically stores external links to other BrainDataFile objects. However, the
          container may also be self-contained.
    """
    @classmethod
    def get_format_specification(cls):
        """
        Get dictionary describing the format.
        """
        return {'datasets': {},
                'groups': {},
                'managed_objects': [{'format_type': 'BrainDataFile', 'optional': True}],
                'attributes': [],
                'group': None,
                'prefix': None,
                'file_prefix': None,
                'file_extension': '.h5',
                'optional': False,
                'description': 'Container file used to organize multiple BrainDataFile objects into a larger ' +
                               'data collection, e.g., to create a collection of recording session or experiments ' +
                               'allowing users to more seamlessly interact with many related files.'}

    def add_object(self, hdf_object, relative_link=True, force_creation=False):
        """
        Add the given BrainDataFile object to the collection.

        :param hdf_object: The object to be added to the collection. This may either be: i) and instance of
                           BrainDataFile, ii) and instance of h5py.Group pointing to a BrainDataFile object,
                           iii) a tuple of two strings (filename, path) or) a dict {'filename':..., 'path':...}
                           describing the location of the file and the path in the hdf5 file. **NOTE**
        :param relative_link: Should we use relative links (i.e., relative to the BrainDataMultiFile where we
                       are adding the link (True) or should we use absolute paths to external files (False). Default
                       value is True.
        :type relative_link: bool
        :param force_creation: Force the creation of the link, even if it is not explicitly permitted by the format
                       specification and/or the link cannot be verified. E.g, we may not be able to always open
                       the linked object to verify that the link actually points to a BrainDataFile managed object
                       or other support object.
        :type force_creation: bool


        :raises ValueError: A ValueError is raised in case that the description of the link is invalid.
        :raises IOError: In case that the link cannot be established

        """
        # 1) Determine the filename and path variable with the settings for the link
        if isinstance(hdf_object, tuple):
            if len(hdf_object) != 2:
                raise ValueError('Invalid description of the link given in hdf_object because len(hdf_object) != 2')
            filename, path = hdf_object
        elif isinstance(hdf_object, dict):
            if 'filename' not in hdf_object or 'path' not in hdf_object:
                raise ValueError('Missing filename and/or path key in hdf_object dict.')
            filename = hdf_object['filename']
            path = hdf_object['path']
        elif isinstance(hdf_object, ManagedObject):
            filename = hdf_object.get_filename(absolute_path=True)
            path = hdf_object.get_h5py().name
        elif isinstance(hdf_object, h5py.Group):
            filename = os.path.abspath(hdf_object.file.filename)
            path = hdf_object.name
        else:
            raise ValueError("Invalid type for hdf_object.")

        # 2) Verify that the filename and path are valid
        if not force_creation:
            if not os.path.exists(filename):
                ValueError("The link was requested to point to an invalid file path: " + filename)
            elif not os.path.isfile(filename):
                ValueError("The requested link does not point to a file: " + filename)

        # 3) Verify that the filename and path follow the format spec, i.e., point to a BrainDataFile object
        if not force_creation:
            temp_file = None
            if isinstance(hdf_object, h5py.Group):
                try:
                    hdf_object = self.get_managed_object(hdf_object)
                except (NameError, ValueError):
                    hdf_object = None
            elif isinstance(hdf_object, tuple) or isinstance(hdf_object, dict):
                try:
                    temp_file = h5py.File(filename, 'r')
                    hdf_object = self.get_managed_object(temp_file[path])
                except (NameError, ValueError):
                    hdf_object = None
                except (KeyError, IOError):
                    raise ValueError('The target object could not be opened.')
            if not isinstance(hdf_object, BrainDataFile):
                raise ValueError('The link did not point to a BrainDataFile as required by the format specification')
            if temp_file:
                temp_file.close()

        # 4) Turn the link into a relative link if necessary or use an absolute link
        if relative_link:
            abs_filename = os.path.abspath(filename)  # Absolute path to the linked file
            my_filename = self.get_filename(absolute_path=True)  # Absolute path to this file
            common_dir_prefix = os.path.commonprefix([os.path.dirname(abs_filename),
                                                      os.path.dirname(my_filename)])
            filename = os.path.relpath(path=abs_filename,
                                       start=common_dir_prefix)
        else:
            filename = os.path.abspath(filename)

        # 5) Construct a name for the link
        # 5.1) Construct a name based on the spec of BrainDataFile
        brain_data_file_spec = BrainDataFile.get_format_specification()
        if brain_data_file_spec['group'] is None:   # TODO Fix name depending on how we handle group names for files
            object_prefix = brain_data_file_spec['prefix']
            object_index = len(BrainDataFile.get_all(parent_group=self.hdf_object))
            link_name = object_prefix + str(object_index)
            # If get_all was not able to open an external link, then our index might be wrong
            if link_name in self.hdf_object.keys():
                existing_keys = self.hdf_object.keys()
                for i in range(object_index, object_index+10000):
                    if object_prefix + str(i) not in existing_keys:
                        link_name = object_prefix + str(i)
                        break
        else:
            link_name = brain_data_file_spec['group']
        # 5.2 Check if the link name is valid
        if link_name in self.hdf_object.keys():
            raise ValueError("Could not create a new link. %s already specified in the file.")
        # 6) Create the external link
        self.hdf_object[link_name] = h5py.ExternalLink(filename=filename,
                                                       path=path)


################################################
# BrainDataInternalData                        #
################################################
class BrainDataData(ManagedGroup):
    """
    Class for management of the `data` group for storage of brain recordings data.

    :ivar hdf_object: See ManagedGroup

    Implicit instance variables (i.e., these are mapped names but not stored explicitly)

    :ivar internal_#: Same as self.internal(#). Usually use internal_0
    :ivar external_#: Same as self.external(#). Usually use external_0

    """
    def __init__(self, hdf_object):
        """
        :param hdf_object: The h5py.Group object managed by the current instance.

        :raises: ValueError in case the given hdf_object does not match the type expected by the class
                 as indicated byt the get_managed_object_type() function.
        """
        super(BrainDataData, self).__init__(hdf_object=hdf_object)

    @classmethod
    def get_format_specification(cls):
        """
        Get dictionary describing the format.
        """
        return {'datasets': {},
                'groups': {},
                'managed_objects': [{'format_type': 'BrainDataInternalData', 'optional': False},
                                    {'format_type': 'BrainDataExternalData', 'optional': False}],
                'attributes': [],
                'group': "data",
                'prefix': None,
                'optional': False,
                'description': 'Managed group for storage of brain data (internal and external).'}

    def populate(self, **kwargs):
        """
        Populate the Brain file with the Data and Descriptors group.
        """
        parent_object = self.hdf_object
        BrainDataInternalData.create(parent_object=parent_object)
        BrainDataExternalData.create(parent_object=parent_object)

    def internal(self, index=0):
        """
        Get the internal managed object with the internal data. In
        principal the current format defines that there should be always
        exactly one internal group. However, multiple could be supported
        so this function is prepared to handle this in case the format should
        change in the future.

        :param index: Optional input parameter (should be always 0 for now)
                      to define which internal group should be retrieved.
                      Currently their should be always exactly one.

        :returns: BrainDataInternalData object or None
        """
        internalobjs = BrainDataInternalData.get_all(self.hdf_object)
        if len(internalobjs) == 0:
            return None
        else:
            return BrainDataInternalData(internalobjs[index])

    def external(self, index=0):
        """
        Get the external managed object with the external data. In
        principal the current format defines that there should be always
        exactly one external group. However, multiple could be supported
        so this function is prepared to handle this in case the format should
        change in the future.

        :param index: Optional input parameter (should be always 0 for now)
                      to define which external group should be retrieved.
                      Currently their should be always exactly one.

        :returns: BrainDataExternalData object or None
        """
        externalobjs = BrainDataExternalData.get_all(self.hdf_object)
        if len(externalobjs) == 0:
            return None
        else:
            return BrainDataExternalData(externalobjs[index])

    def __getattr__(self, item):
        if item.startswith('internal_'):
            indexstr = item.lstrip('internal_')
            if indexstr.isdigit():
                return self.internal(int(indexstr))
            else:
                raise AttributeError("'BrainDataData' has not attribute " + item)
        elif item.startswith('external_'):
            indexstr = item.lstrip('external_')
            if indexstr.isdigit():
                return self.external(int(indexstr))
            else:
                raise AttributeError("'BrainDataData' has not attribute " + item)
        else:
            try:
                return super(BrainDataData, self).__getattribute__(item)
            except AttributeError:
                return super(BrainDataData, self).__getattr__(item)


################################################
# BrainDataInternalData                        #
################################################
class BrainDataInternalData(ManagedGroup):
    """
    Class for management of the 'internal' group for storage of brain recordings data.

    :ivar hdf_object: See ManagedGroup

    Implicit instance variables (i.e., these are mapped names but not stored explicitly)

    :ivar ecog_data_#: Same as self.ecog_data(#) where # is the index of the raw ecog dataset.
    :ivar ecog_data_processed_#: Same as self.ecog_data_processed(#) where # is the index of the processed ecog dataset.

    """
    def __init__(self, hdf_object):
        """
        :param hdf_object: The h5py.Group object managed by the current instance.

        :raises: ValueError in case the given hdf_object does not match the type expected by the class
                 as indicated byt the get_managed_object_type() function.
        """
        super(BrainDataInternalData, self).__init__(hdf_object=hdf_object)

    @classmethod
    def get_format_specification(cls):
        """
        Get dictionary describing the format of the group.
        """
        return {'datasets': {},
                'groups': {},
                'managed_objects': [{'format_type': 'BrainDataECoG', 'optional': True},
                                    {'format_type': 'BrainDataECoGProcessed', 'optional': True}],
                'attributes': [],
                'group': "internal",
                'prefix': None,
                'optional': False,
                'description': 'Managed group for storage of a collection of internal brain data.'}

    def populate(self, **kwargs):
        """
        Populate the 'data' group with all required elements.
        """
        pass

    def num_ecog_data(self):
        """
        Get the number of ecog datasets available.
        """
        return len(BrainDataECoG.get_all(self.hdf_object))

    def num_ecog_data_processed(self):
        """
        Get the number of processed ecog datasets available.
        """
        return len(BrainDataECoGProcessed.get_all(self.hdf_object))

    def ecog_data(self, index=0):
        """
        Get the ecog managed object with the ecog data.

        :param index: Optional input parameter to define the index of the ecog dataset.

        :returns: BrainDataEcoG object or None
        """
        ecogobjs = BrainDataECoG.get_all(self.hdf_object)
        if len(ecogobjs) == 0:
            return None
        else:
            return BrainDataECoG(ecogobjs[index])

    def ecog_data_processed(self, index=0):
        """
        Get the `ecog_processed_` managed object with the processed ecog data.

        :param index: Optional input parameter to define the index of the processed ecog dataset.

        :returns: BrainDataEcoGProcessed object or None
        """
        ecogobjs = BrainDataECoGProcessed.get_all(self.hdf_object)
        if len(ecogobjs) == 0:
            return None
        else:
            return BrainDataECoGProcessed(ecogobjs[index])

    def __getattr__(self, item):
        if item.startswith('ecog_data_processed_'):
            indexstr = item.lstrip('ecog_data_processed_')
            if indexstr.isdigit():
                return self.ecog_data_processed(int(indexstr))
            else:
                raise AttributeError("'BrainDataInternalData' has no attribute " + item)
        elif item.startswith('ecog_data_'):
            indexstr = item.lstrip('ecog_data_')
            if indexstr.isdigit():
                return self.ecog_data(int(indexstr))
            else:
                raise AttributeError("'BrainDataInternalData' has not attribute " + item)
        else:
            try:
                return super(BrainDataInternalData, self).__getattribute__(item)
            except AttributeError:
                return super(BrainDataInternalData, self).__getattr__(item)


################################################
# BrainDataInternalData                        #
################################################
class BrainDataExternalData(ManagedGroup):
    """
    Class for management of the 'external' group for storage of recordings external to the brain.
    """
    def __init__(self, hdf_object):
        """
        :param hdf_object: The h5py.Group object managed by the current instance.

        :raises: ValueError in case the given hdf_object does not match the type expected by the class
                 as indicated byt the get_managed_object_type() function.
        """
        super(BrainDataExternalData, self).__init__(hdf_object=hdf_object)

    @classmethod
    def get_format_specification(cls):
        """
        Get dictionary describing the format of the group.
        """
        return {'datasets': {},
                'groups': {},
                'managed_objects': [],
                'attributes': [],
                'group': "external",
                'prefix': None,
                'optional': False,
                'description': 'Managed group for storage of external data related to the internal brain data.'}

    def populate(self, **kwargs):
        """
        Populate the 'data' group with all required elements.
        """
        pass


################################################
# BrainDataDescriptors                         #
################################################
class BrainDataDescriptors(ManagedGroup):
    """
    Class for management of the `descriptors` group for storage of data descriptions and metadata.

    :ivar hdf_object: See ManagedFile

    Implicit instance variables (i.e., these are mapped names but not stored explicitly)

    :ivar static_#: Same as self.static(#) where # is the index of the static descriptor group.
                    The current format assumes a single static group, i.e., only static_0 is
                    typically valid.
    :ivar dynamic_#: Same as self.dynamic(#) where # is the index of the static descriptor group.
                    The current format assumes a single static group, i.e., only static_0 is
                    typically valid.
    """
    def __init__(self, hdf_object):
        """
        :param hdf_object: The h5py.Group object managed by the current instance.

        :raises: ValueError in case the given hdf_object does not match the type expected by the class
                 as indicated byt the get_managed_object_type() function.
        """
        super(BrainDataDescriptors, self).__init__(hdf_object=hdf_object)

    @classmethod
    def get_format_specification(cls):
        """
        Get dictionary describing the format of the group.
        """
        return {'datasets': {},
                'groups': {},
                'managed_objects': [{'format_type': 'BrainDataStaticDescriptors', 'optional': False},
                                    {'format_type': 'BrainDataDynamicDescriptors', 'optional': False}],
                'attributes': [],
                'group': "descriptors",
                'prefix': None,
                'optional': False,
                'description': 'Managed group for storage of a collection of brain data descriptors.'}

    def populate(self, **kwargs):
        """
        Populate the 'descriptors' group with all required elements.
        """
        parent_object = self.hdf_object
        BrainDataDynamicDescriptors.create(parent_object=parent_object)
        BrainDataStaticDescriptors.create(parent_object=parent_object)

    def static(self, index):
        """
        Get the static descriptors managed object with the static descriptor data.
        In principal the current format defines that there should be always
        exactly one static descriptor group. However, multiple could be supported
        so this function is prepared to handle this in case the format should
        change in the future.

        :param index: Optional input parameter to define the index of static descriptors.
                      Default is index=0.

        :returns: BrainDataEcoGProcessed object or None
        """
        staticobjs = BrainDataStaticDescriptors.get_all(self.hdf_object)
        if len(staticobjs) == 0:
            return None
        else:
            return BrainDataStaticDescriptors(staticobjs[index])

    def dynamic(self, index):
        """
        Get the dynamic descriptors managed object with the dynamic descriptor data.
        In principal the current format defines that there should be always
        exactly one dynamic descriptor group. However, multiple could be supported
        so this function is prepared to handle this in case the format should
        change in the future.

        :param index: Optional input parameter to define the index of dynamic descriptors.
                      Default is index=0.

        :returns: BrainDataEcoGProcessed object or None
        """
        dynamicobjs = BrainDataDynamicDescriptors.get_all(self.hdf_object)
        if len(dynamicobjs) == 0:
            return None
        else:
            return BrainDataDynamicDescriptors(dynamicobjs[index])

    def __getattr__(self, item):
        if item.startswith('static_'):
            indexstr = item.lstrip('static_')
            if indexstr.isdigit():
                return self.static(int(indexstr))
            else:
                raise AttributeError("'BrainDataDescriptors' has not attribute" + item)
        elif item.startswith('dynamic_'):
            indexstr = item.lstrip('dynamic_')
            if indexstr.isdigit():
                return self.dynamic(int(indexstr))
            else:
                raise AttributeError("'BrainDataDescriptors' has not attribute" + item)
        else:
            try:
                return super(BrainDataDescriptors, self).__getattribute__(item)
            except AttributeError:
                return super(BrainDataDescriptors, self).__getattr__(item)


################################################
# BrainDataStaticDescriptors                   #
################################################
class BrainDataStaticDescriptors(ManagedGroup):
    """
    Class for management of the 'static' group for storage of static data descriptors.
    """
    def __init__(self, hdf_object):
        """
        :param hdf_object: The h5py.Group object managed by the current instance.

        :raises: ValueError in case the given hdf_object does not match the type expected by the class
                 as indicated byt the get_managed_object_type() function.
        """
        super(BrainDataStaticDescriptors, self).__init__(hdf_object=hdf_object)

    @classmethod
    def get_format_specification(cls):
        """
        Get dictionary describing the format of the group.
        """
        return {'datasets': {},
                'groups': {},
                'managed_objects': [],
                'attributes': [],
                'group': "static",
                'prefix': None,
                'optional': False,
                'description': 'Managed group for storage of static descriptors.'}

    def populate(self, **kwargs):
        """
        Populate the 'data' group with all required elements.
        """
        pass


################################################
# BrainDataStaticDescriptors                   #
################################################
class BrainDataDynamicDescriptors(ManagedGroup):
    """
    Class for management of the 'dynamic' group for storage of dynamic data descriptors.
    """
    def __init__(self, hdf_object):
        """
        :param hdf_object: The h5py.Group object managed by the current instance.

        :raises: ValueError in case the given hdf_object does not match the type expected by the class
                 as indicated byt the get_managed_object_type() function.
        """
        super(BrainDataDynamicDescriptors, self).__init__(hdf_object=hdf_object)

    @classmethod
    def get_format_specification(cls):
        """
        Get dictionary describing the format of the group.
        """
        return {'datasets': {},
                'groups': {},
                'managed_objects': [],
                'attributes': [],
                'group': "dynamic",
                'prefix': None,
                'optional': False,
                'description': 'Managed group for storage of static descriptors.'}

    def populate(self, **kwargs):
        """
        Populate the 'data' group with all required elements.
        """
        pass


################################################
# BrainDataECoG                                #
################################################
class BrainDataECoG(ManagedGroup):
    """
    Class for management of managed h5py.Group  objects structured to store ECoG brain recording data.

    :ivar hdf_object: See ManagedObject

    """
    def __init__(self, hdf_object):
        """
        :param hdf_object: The h5py.Group or h5py.Dataset object managed by the current instance.

        :raises: ValueError in case the given hdf_object does not match the type expected by the class
                 as indicated byt the get_managed_object_type() function.
        """
        super(BrainDataECoG, self).__init__(hdf_object)

    @classmethod
    def get_format_specification(cls):
        """
        Get dictionary describing the format.
        """
        return {'datasets': {'ecog_data': {'dataset': 'raw_data',             # dataset key is mandatory may be None
                                           'prefix': None,                    # prefix key is mandatory may be None
                                           'optional': False,                 # optional key is mandatory
                                           'primary': True,                   # this is a primary dataset for analysis
                                           # dimensions key is optional. If specified we assume that the number of
                                           # dimensions is fixed and that a scale is specified for all dimensions,
                                           # even if that sepcification is empty, i.e., no fixed scale is given.
                                           # NOTE: if multiple scales are defined for the same axis, then the name
                                           #       for those axis must be the same, while the unit key may differ
                                           #       between scales for the same dimension
                                           'dimensions': [{'name': 'space',     # Mandatory
                                                           'unit': 'id',        # Mandatory
                                                           'optional': False,   # Mandatory
                                                           'dataset': 'electrode_id',  # Mandatory
                                                           'axis': 0},  # Mandatory
                                                          {'name': 'time',
                                                           'unit': 'ns',
                                                           'optional': False,
                                                           'dataset': 'time_axis',
                                                           'axis': 1},  # use name = 'None' for axis without a scale
                                                          {'name': 'space',
                                                           'unit': 'region name',
                                                           'optional': True,
                                                           'dataset': 'anatomy_name',
                                                           'axis': 0},
                                                          {'name': 'space',
                                                           'unit': 'region id',
                                                           'optional': True,
                                                           'dataset': 'anatomy_id',
                                                           'axis': 0}],
                                           'dimensions_fixed': True,
                                           'description': 'Dataset with the ECoG recordings data',  # Mandatory
                                           'attributes': [{'attribute': 'unit',   # Mandatory but may be empty
                                                           'value': 'Volt',  # Mandatory. Use None for user-defined
                                                           'prefix': None,        # Mandatory
                                                           'optional': False}]},   # Mandatory
                             'sampling_rate': {'dataset': 'sampling_rate',
                                               'prefix': None,
                                               'optional': False,
                                               'attributes': [{'attribute': 'unit',
                                                               'value': 'KHz',
                                                               'prefix': None,
                                                               'optional': False}],
                                               'description': 'Sampling rate in KHz'},
                             'layout': {'dataset': 'layout',
                                        'prefix': None,
                                        'optional': True,
                                        'dimensions': [],
                                        'attributes': [],
                                        'description': 'The physical layout of the electrodes.'}},
                'groups': {},           # Mandatory
                'managed_objects': [{'format_type': 'AnnotationDataGroup',
                                     'optional': True}],  # e.g.,{'format_type': 'BrainDataECoG', 'optional': True}
                'attributes': [],  # e.g, [{'attribute': 'unit', 'value': KHz, 'prefix': None, 'optional': False}]
                'group': None,          # Mandatory (use 'dataset' in case of a managed dataset
                'prefix': "ecog_data_",
                'optional': False,
                'description': 'Managed group for storage of raw ECoG recordings.'}
                # For file descriptions set the group="/" . If prefix is given it is assumed that all
                # names file of that type must start with prefix.

    def populate(self,
                 ecog_data=None,
                 ecog_data_shape=None,
                 ecog_data_type=None,
                 sampling_rate=16,
                 start_time=0,
                 layout=None,
                 anatomy_names=None,
                 anatomy_ids=None,
                 format_spec=None):
        """
        The populate method is called by the create method after the basic common setup is complete.
        The function should be used to populate the managed object (e.g., add dimensions to a datadset or
        add required datasets to a group. The populate method is passed the kwargs handed to the create
        method.

        :param ecog_data: The neurallogical recordings data to be written. If present then ecog_data_shape
                           and ecog_data_dtype are ignored.
        :type ecog_data: 2D numpy array of [#channels, #samplesPerChannel]
        :param ecog_data_shape: Shape of the neural data set. Required if ecog_data is not present. If
                           ecog_data_shape is ignored.
        :type ecog_data_shape: 2D tuple indicating the number of recordings and samples per recording.
        :param ecog_data_type: Datatype of the neural data set. Required if ecog_data is not present. If
                           ecog_data_shape is ignored.
        :type ecog_data_type: numpy.dtype or string indicating the dtype
        :param sampling_rate: Sampling rate in KHz
        :type sampling_rate: int or float
        :param start_time: Start time of the recording in time since epoch (default behavior by time.time())
                           The Unix epoch (or Unix time or POSIX time or Unix timestamp) is the number of seconds
                           that have elapsed since January 1, 1970 (midnight UTC/GMT), not counting leap seconds (
                           in ISO 8601: 1970-01-01T00:00:00Z). Literally speaking the epoch is Unix time 0
                           (midnight 1/1/1970), but 'epoch' is often used as a synonym for 'Unix time'.
        :type start_time: int64
        :param layout: The spatial layout of the ecog_data.
        :type layout: numpy array of ints describing for each channel index its location in the layout matrix.
                      Use -1 to indicate unoccupied channels. This strategy allows complex layouts to be described
                      via a rectangular matrix.
        :param anatomy_names: 1D array of strings with the name of the region an electrode is located in.
        :param anatomy_ids: 1D array of ints with the integer id's of the regions the electrode. Note,
                            if anatomy_names is given and anatomy_ids is empty, then the array will be
                            autogenerated by creating an index for the unique values in the anatomy_names array.
        :param format_spec: Format specification. Usually this is None, indicating that the specification defined
                            by self.get_format_specification should be used. In some cases, derived classes will
                            expand the base class specification or use the same specification, simply with
                            different names and values. To allow the base classes to reuse the populate function,
                            the function allows a different format specification to be provided as input.
                            CAUTION! The format_spec must be compliant with the structure described by the spec
                            for this class in order for the populate function to work correctly.

        :raises: ValueError is raised in case that the object cannot be populated.
        """

        # Create the neural dataset
        if not format_spec:
            format_spec = self.get_format_specification()
        # Set the format specification attribute
        self.hdf_object.attrs[ManagedObject.format_spec_attribute_name] = json.dumps(format_spec)
        data_spec = format_spec['datasets']
        if ecog_data is not None:
            ecog_data_shape = ecog_data.shape
        # Check that the number of dimensions of the dataset are valid
        expected_num_dims = self.get_num_dimensions_from_dataset_specification(data_spec)
        if expected_num_dims is not None:
            if expected_num_dims[0] > len(ecog_data.shape):
                raise ValueError('ecog_data must have at least %s dimensions found %s' %
                                 (str(expected_num_dims[0]), str(len(ecog_data.shape))))
            if expected_num_dims[1] < len(ecog_data.shape):
                raise ValueError('ecog_data must have at most %s dimensions found %s' %
                                 (str(expected_num_dims[1]), str(len(ecog_data.shape))))
        # Create the maximum shape for the dataset
        ecog_data_maxshape = [ecog_data_shape[0], None]
        for additional_dim in range(2, len(ecog_data_shape)):
            ecog_data_maxshape.append(ecog_data_shape[additional_dim])
        ecog_data_maxshape = tuple(ecog_data_maxshape)
        # Generate the dataset either by setting the data or generating an empty dataset of the given shape
        if ecog_data is not None:
            ecog_dataset = self.hdf_object.create_dataset(name=data_spec['ecog_data']['dataset'],
                                                          data=ecog_data,
                                                          dtype=ecog_data.dtype,
                                                          chunks=True,
                                                          compression='gzip',
                                                          compression_opts=2,
                                                          maxshape=ecog_data_maxshape,
                                                          fillvalue=np.asarray(0, dtype=ecog_data.dtype))
        elif ecog_data_shape is not None and ecog_data_type is not None:
            ecog_dataset = self.hdf_object.require_dataset(name=data_spec['ecog_data']['dataset'],
                                                           shape=ecog_data_shape,
                                                           dtype=ecog_data_type,
                                                           chunks=True,
                                                           compression='gzip',
                                                           compression_opts=2,
                                                           maxshape=ecog_data_maxshape,
                                                           fillvalue=0)
        else:
            raise ValueError("If ecog_data is not set then ecog_data_shape and ecog_data_dtype are required.")

        # Create the basic attributes
        ecog_dataset.attrs[ManagedObject.description_attribute_name] = data_spec['ecog_data']['description']
        ecog_data_attr_spec = data_spec['ecog_data']['attributes'][0]
        if ecog_data_attr_spec['value'] is not None:
            ecog_dataset.attrs[ecog_data_attr_spec['attribute']] = ecog_data_attr_spec['value']
        else:
            ecog_dataset.attrs[ecog_data_attr_spec['attribute']] = ""

        # Create the time axis
        sample_period = (100. / sampling_rate)  # Sample period in ns
        stop_time = start_time + sample_period*ecog_data_shape[1]
        timedata = np.arange(start_time, stop_time, sample_period)
        time_dataset = self.hdf_object.create_dataset(name=data_spec['ecog_data']['dimensions'][1]['dataset'],
                                                      data=timedata)

        # Create spatial coordinate map
        electrode_id_dataset = self.hdf_object.create_dataset(name=data_spec['ecog_data']['dimensions'][0]['dataset'],
                                                              data=np.arange(ecog_dataset.shape[0]))

        # Save the units for the different dimensions as the label of the dimensions
        eid_axis_index = data_spec['ecog_data']['dimensions'][0]['axis']
        time_axis_index = data_spec['ecog_data']['dimensions'][1]['axis']
        ecog_dataset.dims[eid_axis_index].label = data_spec['ecog_data']['dimensions'][0]['name']
        ecog_dataset.dims[time_axis_index].label = data_spec['ecog_data']['dimensions'][1]['name']
        # Create the scales for the different dimensions from the corresponding datasets created above
        ecog_dataset.dims.create_scale(electrode_id_dataset, data_spec['ecog_data']['dimensions'][0]['unit'])
        ecog_dataset.dims.create_scale(time_dataset, data_spec['ecog_data']['dimensions'][1]['unit'])
        # Attach the scales to the dataset
        ecog_dataset.dims[eid_axis_index].attach_scale(electrode_id_dataset)
        ecog_dataset.dims[time_axis_index].attach_scale(time_dataset)

        # Save the sampling rate
        self.hdf_object[data_spec['sampling_rate']['dataset']] = sampling_rate
        sr_dataset = self.hdf_object[data_spec['sampling_rate']['dataset']]
        sr_dataset.attrs[ManagedObject.description_attribute_name] = data_spec['sampling_rate']['description']
        sr_unit_attr_spec = data_spec['sampling_rate']['attributes'][0]
        sr_dataset.attrs[sr_unit_attr_spec['attribute']] = sr_unit_attr_spec['value']

        # Save the layout dataset if available
        if layout is not None:
            self.hdf_object[data_spec['layout']['dataset']] = layout
            layout_dataset = self.hdf_object[data_spec['layout']['dataset']]
            layout_dataset.attrs[ManagedObject.description_attribute_name] = data_spec['layout']['description']

        # Create/save the annotations
        my_annotations = AnnotationDataGroup.create(parent_object=self.hdf_object,
                                                    data_object=ecog_dataset,
                                                    collection_description='anatomy')

        # Initialize the anatomies ids array if needed. Note annotations need to be initialized first
        # as the anatomy will be stored as both dimension scales and as annotations.
        self.anatomy(anatomy_names=anatomy_names,
                     anatomy_ids=anatomy_ids,
                     format_spec=format_spec,
                     annotation_data_group=my_annotations)

    def dims(self, get_hdf=True, axis=None):
        """
        Get the h5py Dimension scales associated with the ecog dataset

        :param get_hdf: Get the h5py.DimensionScale associated with the
                        ecog dataset if set to True (Default). If set to
                        False a list of lists of  dicts with a summary of
                        the dimensions is returned instead which is formatted as follows:
                        [axis_index][scale_index]. Each dict then
                        contains the name, unit, dataset, and axis information
                        for the dimensions scale.
        """
        dimensions = self.ecog_data().dims
        if get_hdf:
            if axis is None:
                return dimensions
            else:
                return dimensions[axis]
        else:
            dim_list = [[] for _ in range(len(self.ecog_data().shape))]
            for dimid, dim in enumerate(dimensions):
                for subscale in dim.keys():
                    dim_list[dimid].append({'name': dim.label,
                                            'unit': subscale,
                                            'dataset': dim[subscale],
                                            'axis': dimid})
            if axis is None:
                return dim_list
            else:
                return dim_list[axis]

    def anatomy(self,
                anatomy_names=None,
                anatomy_ids=None,
                format_spec=None,
                annotation_data_group=None,
                get_hdf=True):
        """
        Get the anatomy information associated with the dataset.

        :param anatomy_names: 1D array of strings with the name of the region an electrode is located in.
                            By setting this parameter we can add an anatomical description to a dataset
                            if it does not has one.
        :param anatomy_ids: 1D array of ints with the integer id's of the regions the electrode. Note,
                            if anatomy_names is given and anatomy_ids is empty, then the array will be
                            autogenerated by creating an index for the unique values in the anatomy_names array.
                            By setting this parameter we can add an anatomical description to a dataset
                            if it does not has one.
        :param format_spec: Format specification. Usually this is None, indicating that the specification defined
                            by self.get_format_specification should be used. In some cases, derived classes will
                            expand the base class specification or use the same specification, simply with
                            different names and values. To allow the base classes to reuse the populate function,
                            the function allows a different format specification to be provided as input.
                            CAUTION! The format_spec most be compliant with the structure described by the spec
                            for this class in order for the populate function to work correctly.
        :param annotation_data_group: Optional AnnotationDataGroup object used to store anatomy description as
                            annotations. This is only relevant when either the anatomy_names and/or anatomy_ids
                            parameter is specified. If the annotation_data_group is given then the anatomy will
                            automatically be transformed to annotations and added to the given annotations object.
        :type annotation_data_group: AnnotationDataGroup
        :param get_hdf: Get the value in numpy format (set to False) or the
                         h5py dataset for the layout data (set to True).
                         Default value is True.
        :type get_hdf: bool

        :raises KeyError: A KeyError is generated if a new anatomy_names array or anatomy_ids array is
                          given but a prior description already exists.

        :returns: 1D Numpy array indicating for each recording the name of the region. The
                  function also returns a second 1D numpy indicating for each recording the index
                  of the region the electrode was located in. Note, if any of the anatomy
                  arrays are missing, then None will be returned instead of the anatomy array.
        """
        # Create the anatomy data storage if possible
        if anatomy_names is not None or anatomy_ids is not None:
            if self.has_anatomy():
                raise KeyError("An anatomy description is already defined for the dataset.")
            # Initialize the anatomies ids array if needed
            if anatomy_names is not None and anatomy_ids is None:
                unique_names = np.unique(anatomy_names)
                anatomy_ids = np.zeros(shape=(anatomy_names.size, ),
                                       dtype='int')
                for aindex, aname in enumerate(unique_names):
                    anatomy_ids[anatomy_names == aname] = aindex

            # Get the format specification and ecog dataset
            if not format_spec:
                format_spec = self.get_format_specification()
            data_spec = format_spec['datasets']
            ecog_dataset = self.ecog_data()

            # Create the anatomy_names scale
            if anatomy_names is not None:
                anatomy_axis_index = data_spec['ecog_data']['dimensions'][2]['axis']
                anatomy_name_dataset = self.hdf_object.create_dataset(
                    name=data_spec['ecog_data']['dimensions'][2]['dataset'],
                    data=anatomy_names)
                ecog_dataset.dims.create_scale(anatomy_name_dataset, data_spec['ecog_data']['dimensions'][2]['unit'])
                ecog_dataset.dims[anatomy_axis_index].attach_scale(anatomy_name_dataset)

            # Create the anatomy_ids scale
            if anatomy_ids is not None:
                anatomy_axis_index = data_spec['ecog_data']['dimensions'][3]['axis']
                anatomy_ids_dataset = self.hdf_object.create_dataset(
                    name=data_spec['ecog_data']['dimensions'][3]['dataset'],
                    data=anatomy_ids)
                ecog_dataset.dims.create_scale(anatomy_ids_dataset, data_spec['ecog_data']['dimensions'][3]['unit'])
                ecog_dataset.dims[anatomy_axis_index].attach_scale(anatomy_ids_dataset)

            # Create an annotation for each unique anatomical description
            if annotation_data_group is not None:
                if anatomy_names is not None:
                    anatomy_array = anatomy_names
                else:
                    anatomy_array = anatomy_ids
                unique_anatomy_names = np.unique(anatomy_array)
                for aname in unique_anatomy_names:
                    selection_object = DataSelection(data_object=annotation_data_group.data_object)
                    selection_object[0, :] = anatomy_array == aname  # Equivalent to: selection_object['space'. :] = ..
                    anno = Annotation(annotation_type='anatomy',
                                      description=unicode(aname),
                                      data_selection=selection_object)
                    annotation_data_group.add_annotation(annotation=anno)
            self.hdf_object.file.flush()

        # Retrieve the anatomy data and return it
        anatomy_names = None
        anatomy_ids = None
        dimension_spec = self.get_format_specification()['datasets']['ecog_data']['dimensions']
        anatomy_names_spec = dimension_spec[2]
        anatomy_ids_spec = dimension_spec[3]
        space_axes = self.ecog_data().dims[anatomy_names_spec['axis']]
        if anatomy_names_spec['unit'] in space_axes.keys():
            anatomy_names = space_axes[anatomy_names_spec['unit']]
            if not get_hdf:
                anatomy_names = anatomy_names[:]
        if anatomy_ids_spec['unit'] in space_axes.keys():
            anatomy_ids = space_axes[anatomy_ids_spec['unit']]
            if not get_hdf:
                anatomy_ids = anatomy_ids[:]
        return anatomy_names, anatomy_ids

    def has_anatomy(self):
        """
        Check whether an anatomy description is available for the ECoG dataset.

        :returns: Boolean indicating whether an anatomy description by name and or id is available.
        """
        anames, aids = self.anatomy()
        return (anames is not None) or (aids is not None)

    def ecog_data(self, get_hdf=True):
        """
        Get the h5py.Dataset object of the ecog data.

        For get_hdf==True this is the same as self['raw_data'] and self['ecog_data'].

        For get_hdf==False this is the same as self['raw_data'][:] and self['ecog_data'][:]

        :param get_hdf: Get the value in numpy format (set to False) or the
                         h5py dataset for the sampling rate (set to True).
                         Default value is True.

        :returns: h5py.Dataset if get_hdf is True. numpy array of the full data
                  if get_hdf is False. None in case that the ecog data should be
                  missing
        """
        format_spec = self.get_format_specification()
        ecog_data_name = format_spec['datasets']['ecog_data']['dataset']
        try:
            if get_hdf:
                return self[ecog_data_name]
            else:
                return self[ecog_data_name][:]
        except KeyError:
            return None

    def sampling_rate(self, get_hdf=True):
        """
        Get the h5py.Dataset object of the sampling rate data. Same as
        self['sampling_rate']

        :param get_hdf: Get the value in numpy format (set to False) or the
                         h5py dataset for the sampling rate (set to True).
                         Default value is True.

        :returns: h5py.Dataset if get_hdf is True. Float if get_hdf is False.
                  None in case the sampling_rate data should be missing.

        """
        format_spec = self.get_format_specification()
        sr_data_name = format_spec['datasets']['sampling_rate']['dataset']
        try:
            if get_hdf:
                return self[sr_data_name]
            else:
                return self[sr_data_name][()]
        except KeyError:
            return None

    def layout(self, get_hdf=True, new_layout=None):
        """
        Get the h5py.Dataset of the layout of the ecog_data. Same as self['layout']

        :param get_hdf: Get the value in numpy format (set to False) or the
                         h5py dataset for the layout data (set to True).
                         Default value is True.
        :param new_layout: If a layout dataset has not been defined previously then
                         we can generate a new layout by providing an according
                         numpy data array (or h5py.Dataset) to this parameter.
        :returns: h5py.Dataset if get_hdf is True. Numpy array if get_hdf is False.
                  None in case the layout data does not exist.

        """
        format_spec = self.get_format_specification()
        layout_data_name = format_spec['datasets']['layout']['dataset']
        if new_layout is not None:
            if layout_data_name not in self.hdf_object.keys():
                self.hdf_object[layout_data_name] = new_layout[:]
            else:
                raise KeyError('A layout is already defined for the dataset. You may update the h5py object directly.')
        try:
            if get_hdf:
                return self[layout_data_name]
            else:
                return self[layout_data_name][:]
        except KeyError:
            return None

    def has_layout(self):
        """
        Check whether a layout description is available for the ECoG dataset.

        :returns: Boolean indicating whether a layout dataset is available.
        """
        return self.layout() is not None

    def annotations(self, index=None):
        """
        Get the requested annotation object.

        :param index: The index of the annotation object to be retrieved.

        :returns: If index is specified to a value >=0 then a single AnnotationDataGroup object is returned
                  (or None if the index is invalid). Otherwise a list of all AnnoationDataGroup object is returned.
        """
        annotation_spec = AnnotationDataGroup.get_format_specification()
        annotation_prefix = annotation_spec['prefix']
        if index >= 0:
            annotation_name = annotation_prefix + str(index)
            if annotation_name in self.hdf_object.keys():
                return self.get_managed_object(hdf_object=self.hdf_object[annotation_name])
            else:
                return None
        else:
            return AnnotationDataGroup.get_all(parent_group=self.hdf_object,
                                               get_h5py=False)

    def num_annotations(self):
        """
        Get the number of annotation objects associated with the ECoG data.
        """
        return len(self.annotations())

    def has_annotations(self):
        """
        Check whether annotations exists for the ECoG dataset.
        """
        return self.num_annotations() > 0

    def __getitem__(self, item):
        """
        Read data.
        """
        try:
            if item == 'ecog_data':
                format_spec = self.get_format_specification()
                item = format_spec['datasets']['ecog_data']['dataset']
            return self.hdf_object[item]
        except (KeyError, AttributeError):
            if isinstance(item, str) or isinstance(item, unicode):
                raise KeyError('Invalid key given')
            else:
                format_spec = self.get_format_specification()
                ecog_data_name = format_spec['datasets']['ecog_data']['dataset']
                return self.hdf_object[ecog_data_name][item]

    def __setitem__(self, key, value):
        """
        Write data.
        """
        #Check whether the given key has managed name

        #if key not in self.allowed_data_keys:
        #    if self.enforce_format:
        #        raise KeyError('None format compliant key specified for data write.')
        #    else:
        #        warnings.warn('None format compliant key specified for data write.' +
        #                      'The dataset will be added but the resulting file is ' +
        #                      'no longer fully compliant with the Brain Format.')
        if isinstance(key, tuple) or isinstance(key, slice):
            format_spec = self.get_format_specification()
            ecog_data_name = format_spec['datasets']['ecog_data']['dataset']
            self.hdf_object[ecog_data_name][key] = value
        elif isinstance(key, str) or isinstance(key, unicode):
            if key == 'ecog_data':
                key = self.get_format_specification()['datasets']['ecog_data']['dataset']
            self.hdf_object[key] = value


################################################
# BrainDataECoGProcessed                       #
################################################
class BrainDataECoGProcessed(BrainDataECoG):
    """
    Class for management of h5py.Group objects structured to store processed ECoG brain recording data.

    :ivar hdf_object: See ManagedObject and BrainDataECoG.

    """
    def __init__(self, hdf_object):
        """
        :param hdf_object: The h5py.Group or h5py.Dataset object managed by the current instance.

        :raises: ValueError in case the given hdf_object does not match the type expected by the class
                 as indicated byt the get_managed_object_type() function.
        """
        super(BrainDataECoGProcessed, self).__init__(hdf_object)

    @classmethod
    def get_format_specification(cls):
        """
        Get dictionary describing the format.

        This function adapts the specification of the BrainDataECoG class.

        :returns: Dictionary with the format specification.

        """
        # Update the spec from the base class, i.e., BrainDataECoG.
        spec = super(BrainDataECoGProcessed, cls).get_format_specification()
        # Update the prefix for the group
        spec['prefix'] = 'ecog_data_processed_'
        # Set the description and name for the main dataset
        spec['datasets']['ecog_data']['dataset'] = 'processed_data'
        spec['description'] = 'Managed group for storage of processed ECoG recordings.'
        # Make the value of 'units' attribute to None indicating that the value is user-defined.
        spec['datasets']['ecog_data']['attributes'][0]['value'] = None
        # Add optional band attribute to indicate the bands that have been processed
        spec['datasets']['ecog_data']['attributes'].append({'attribute': 'original_name',
                                                            'value': None,
                                                            'prefix': None,
                                                            'optional': True})
        # Add dimension for frequency bands in the processed data
        spec['datasets']['ecog_data']['dimensions'].append({'name': 'frequency bands',
                                                            'unit': 'KHz',
                                                            'optional': True,
                                                            'dataset': 'frequency_bands',
                                                            'axis': 2})
        return spec

    def populate(self,
                 ecog_data=None,
                 ecog_data_shape=None,
                 ecog_data_type=None,
                 ecog_data_units=None,
                 sampling_rate=16,
                 start_time=0,
                 layout=None,
                 anatomy_names=None,
                 anatomy_ids=None,
                 original_name=None,
                 frequency_bands=None):
        """
        The populate method is called by the create method after the basic common setup is complete.
        The function should be used to populate the managed object (e.g., add dimensions to a datadset or
        add required datasets to a group. The populate method is passed the kwargs handed to the create
        method.

        :param ecog_data: The neurallogical recordings data to be written. If present then ecog_data_shape
                           and ecog_data_dtype are ignored.
        :type ecog_data: 2D numpy array of [#channels, #samplesPerChannel]
        :param ecog_data_shape: Shape of the neural data set. Required if ecog_data is not present. If
                           ecog_data_shape is ignored.
        :type ecog_data_shape: 2D tuple indicating the number of recordings and samples per recording.
        :param ecog_data_type: Datatype of the neural data set. Required if ecog_data is not present. If
                           ecog_data_shape is ignored.
        :type ecog_data_type: numpy.dtype or string indicating the dtype
        :param ecog_data_units: String indicating the units used for the processed ECoG recordings.
        :type ecog_data_units: String
        :param sampling_rate: Sampling rate in KHz
        :type sampling_rate: int or float
        :param start_time: Start time of the recording in time since epoch (default behavior by time.time())
                           The Unix epoch (or Unix time or POSIX time or Unix timestamp) is the number of seconds
                           that have elapsed since January 1, 1970 (midnight UTC/GMT), not counting leap seconds (
                           in ISO 8601: 1970-01-01T00:00:00Z). Literally speaking the epoch is Unix time 0
                           (midnight 1/1/1970), but 'epoch' is often used as a synonym for 'Unix time'.
        :type start_time: int64
        :param layout: The spatial layout of the ecog_data
        :type layout: numpy array of ints describing for each channel index its location in the layout matrix.
        :param anatomy_names: 1D array of strings with the name of the region an electrode is located in.
        :param anatomy_ids: 1D array of ints with the integer id's of the regions the electrode. Note,
                            if anatomy_names is given and anatomy_ids is empty, then the array will be
        :param original_name: Original name of the dataset as specified by the user
        :type original_name: String
        :param frequency_bands: Numpy array indicating the center of the frequency bands stored in the
                                third dimension of the processed data array. May be None if the bands are not known.
        :type frequency_bands: 1D Numpy array fo length ecog_data.shape[2] (or ecog_data_shape[2], whichever applies).
                               Default value is None, in which case the dimensions scale for the third dimension is
                               omitted (as it is optional in the spec).

        :raises: ValueError is raised in case that the object cannot be populated.
        """
        spec = self.get_format_specification()
        super(BrainDataECoGProcessed, self).populate(ecog_data=ecog_data,
                                                     ecog_data_shape=ecog_data_shape,
                                                     ecog_data_type=ecog_data_type,
                                                     sampling_rate=sampling_rate,
                                                     start_time=start_time,
                                                     layout=layout,
                                                     anatomy_names=anatomy_names,
                                                     anatomy_ids=anatomy_ids,
                                                     format_spec=spec)
        ecog_dataset = self.hdf_object[spec['datasets']['ecog_data']['dataset']]
        # Define the units attribute (emit warning if not given as the attribute should not be optional)
        unit_attr_spec = spec['datasets']['ecog_data']['attributes'][0]
        if ecog_data_units is not None:
            ecog_dataset.attrs[unit_attr_spec['attribute']] = ecog_data_units
        else:
            warnings.warn('Units for processed ECoG data are empty.')
        # Define the original name attribute (do not warn as it is optional
        if original_name is not None:
            orig_name_attr_spec = spec['datasets']['ecog_data']['attributes'][1]
            ecog_dataset.attrs[orig_name_attr_spec['attribute']] = original_name

        # Add the optional dimension scale for the frequency bands
        band_dim_spec = spec['datasets']['ecog_data']['dimensions'][-1]
        if len(ecog_dataset.shape) > band_dim_spec['axis']:
            ecog_dataset.dims.label = band_dim_spec['name']
            if frequency_bands is not None:
                frequency_bands_dataset = self.hdf_object.create_dataset(name=band_dim_spec['dataset'],
                                                                         data=frequency_bands)
                ecog_dataset.dims.create_scale(frequency_bands_dataset, band_dim_spec['unit'])
                ecog_dataset.dims[band_dim_spec['axis']].attach_scale(frequency_bands_dataset)

    def original_name(self):
        """
        Get the original name specificed by the user for the dataset.

        :return: String indicating the original name of the dataset
                 specified by the user or None in case no user-defined
                 name was specified.
        """
        spec = self.get_format_specification()
        orig_name_attr_spec = spec['datasets']['ecog_data']['attributes'][1]
        orig_name_attr = orig_name_attr_spec['attribute']
        ecog_dataset = self.hdf_object[spec['datasets']['ecog_data']['dataset']]
        if orig_name_attr in ecog_dataset.attrs:
            return ecog_dataset.attrs[orig_name_attr]
        else:
            return None
